#ifndef POT_HH__
#define POT_HH__

#include <map>
#include <vector>
#include <string>
#include <stdexcept>
#include <utils.hh>

namespace libhw
{
  class pot
  {
  public:
    typedef std::multimap<std::string, std::string> attrs_t;
    typedef attrs_t::value_type attr_type;
    typedef std::map<std::string, attrs_t> tree_t;

    struct not_found: public std::runtime_error
    {
      not_found(const std::string& s)
	: std::runtime_error(s)
      {}
    };
  protected:
    tree_t tree_;

  public:

    template <typename Pot>
    class base_node
    {
      Pot& p_;
      std::string name_;

      base_node(Pot& p, const std::string& name)
	: p_(p), name_(name)
      {}

    public:
      template <typename OtherPot>
      base_node(const base_node<OtherPot>& o)
	: p_(o.pot()), name_(o.name())
      {}

      bool empty(void) const
      {
	pot::tree_t::const_iterator i =
	  p_.tree_.find(name_);
	if(i == p_.tree_.end()) return true;

	return i->second.empty();
      }


      Pot& pot(void) const { return p_; }
      const std::string& name(void) const { return name_; }

      const base_node& remove(const std::string& key) const
      {
	pot::attrs_t& attrs = p_.insert_leaf(name_)->second;
        pot::attrs_t::iterator j = attrs.find(key);
	while (j != attrs.end() && j->first == key)
	  {
	    attrs.erase(j++);
	  }
	return *this;
      }

      const base_node&
      operator()(const std::string& key,
		 const std::string& value, bool overwrite = false) const
      {
	pot::attrs_t& attrs =
	  p_.insert_leaf(name_)->second;
	if (overwrite)
	    {
	       pot::attrs_t::iterator j = attrs.find(key);
		while (j != attrs.end() && j->first == key)
		  {
		    attrs.erase(j++);
		  }	
	    }
	  attrs.insert(
		       std::make_pair(key, value));
	return *this;
      }

      const base_node&
      operator()(const std::pair<const std::string, std::string> & pair) const
      {
	return (*this)(pair.first, pair.second);
      }

      std::string
      operator()(const std::string& key) const
      {
	pot::tree_t::const_iterator i =
	  p_.tree_.find(name_);
	if(i == p_.tree_.end()) return "";

	const pot::attrs_t& attrs = i->second;
	pot::attrs_t::const_iterator j =
	  attrs.find(key);
	if(j == attrs.end()) return "";
	return j->second;
      }

      base_node
      operator[](const std::string& leaf) const
      {
	return get(leaf);
      }

      base_node
      get(const std::string& leaf) const
      {
	return p_.get(name_ + "/" + leaf);
      }
      bool
      exists_leaf(const std::string& leaf) const
      {
	return p_.exists(name_ + "/" + leaf);
      }

      bool
      exists_key(const std::string& key) const
      {
	pot::tree_t::const_iterator i =
	  p_.tree_.find(name_);
	if(i == p_.tree_.end()) return false;

	const pot::attrs_t& attrs = i->second;
	return attrs.find(key) != attrs.end();
      }

      template <typename Actor>
      Actor
      for_each_leaf(Actor a) const;

      template <typename Actor>
      Actor
      for_each_branch(Actor a) const;

      template <typename Actor>
      Actor
      for_each_key(Actor a) const;

      friend class pot;
    };

  public:
    friend class node;
    friend class const_node;
    typedef base_node<pot> node;
    typedef base_node<const pot> const_node;

  private:
    tree_t::iterator
    insert_leaf(const std::string& real_leaf);

    bool
    exists_leaf(const std::string& real_leaf) const;

  public:
    node
    insert(const std::string& leaf);

    node
    get(const std::string& leaf);

    const_node
    get(const std::string& leaf) const;

    bool
    exists(const std::string& leaf) const;

    node
    operator[](const std::string& leaf) { return get(leaf); }

    const_node
    operator[](const std::string& leaf) const {return get(leaf);}

    void
    clear(void) { tree_.clear(); }

  };

  template <typename Pot> template <typename Actor>
  Actor
  pot::base_node<Pot>::for_each_leaf(Actor a) const
  {
    std::vector<std::string> was;
    std::string real_prefix = alt::ensure_slash(name_);
    pot::tree_t::const_iterator i =
      p_.tree_.lower_bound(real_prefix);

    while(i != p_.tree_.end() &&
	  alt::starts_with(i->first, real_prefix))
      {
	std::string branch =
	  i->first.substr(real_prefix.length());

	std::string leaf = branch.substr(0, branch.find("/"));
	if(!leaf.empty() &&
	   std::find(was.begin(), was.end(),
		     leaf) == was.end())
	  {
	    const_node n = get(leaf);
	    a(leaf, n);
	    was.push_back(leaf);
	  }
	++i;
      }
    return a;
  }

  template <typename Pot> template <typename Actor>
  Actor
  pot::base_node<Pot>::for_each_branch(Actor a) const
  {
    std::string real_prefix = alt::ensure_slash(name_);
    pot::tree_t::const_iterator i =
      p_.tree_.lower_bound(real_prefix);
      

    while(i != p_.tree_.end() &&
	  alt::starts_with(i->first, real_prefix))
      {
	if(!i->second.empty())
	  {
	    a(i->first,i->second);
	  }
	++i;
      }
    return a;
  }

  template <typename Pot> template <typename Actor>
  Actor
  pot::base_node<Pot>::for_each_key(Actor a) const
  {
    pot::tree_t::const_iterator i =
      p_.tree_.find(alt::ensure_slash(name_));
    if(i != p_.tree_.end())
      {
	const attrs_t& attrs = i->second;
	std::for_each(attrs.begin(), attrs.end(), a);
      }
    return a;
  }

  inline const pot::node&
  insert_key(const pot::node& n,
	     const std::string& key, const std::string& value)
  {
    return n(key, value);
  }
}

#endif // POT_HH__

